/*
 * Copyright (c) 2014 by David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 *
 * Generic input routine.
 */

#include <string.h>

#include "dpy_priv.h"

#define	ISERASE(ch)	(((ch) == '\b') || ((ch) == '\177'))
#define	ISBLANK(ch)	(((ch) == ' ') || ((ch) == '\t'))


/*
 * The default read routine for dpyread.
 */
static int readLine(int);


/*
 * Read some input while possibly showing it in the current window.
 * If the prompt string is NULL, then editing is performed without
 * any windowing activity (useful when reading commands from scripts).
 * Otherwise, the prompt is shown in the window along with any input.
 * The given routine is called for each character, with an argument
 * which is the previous character (or -1 on the first call).
 * The routine returns the next input character, or -1 to stop reading.
 * A null routine defaults to one which reads until an end of line
 * (which is not included in the buffer).  A trailing null character
 * is inserted in the buffer.  The buffer may contain initial data which
 * will initially appear in the window, and which can be edited by the
 * user.  InitCount specifies the number of initial characters in the
 * buffer, so that a value of zero indicates that there is no such data.
 * Scrolling of the window is automatically performed when necessary.
 * Editing of the input is handled.  If the buffer fills up, the user
 * is warned with beeps and further input is ignored.
 * Returns number of bytes of data read.
 */
int
dpyread(const char * prompt, int (*routine)(int),
	char * buf, int count, int initCount)
{
	WINDOW * wp = _dpy_window_;

	int ch;			/* character which was read */
	char * bp;		/* current buffer pointer location */
	char * endBp;		/* end of buffer */
	BOOL isRedrawNeeded;	/* need to redisplay input */
	BOOL isOldScroll;	/* old scrolling flag */
	int promptLen;		/* length of prompt string */

	promptLen = 0;

	if (prompt)
		promptLen = strlen(prompt);

	if (routine == 0)
		routine = readLine;

	bp = buf;
	endBp = bp + count - 1;
	isRedrawNeeded = TRUE;
	ch = -1;
	isOldScroll = wp->isScroll;
	wp->isScroll = TRUE;

	if (initCount > 0)
	{
		if (initCount >= count)
			initCount = count - 1;

		bp = &buf[initCount];
	}

	for (;;)
	{
		if (prompt && isRedrawNeeded)
		{
			/*
			 * Recompute window data.
			 */
			isRedrawNeeded = FALSE;
			dpyhome();
			dpywrite(prompt, promptLen);
			dpywrite(buf, bp - buf);
			dpyclrwindow();
		}

		if (prompt)
			dpyupdate();

		ch = routine(ch);

		if (ch < 0)
		{
			/*
			 * End of file.
			 */
			wp->isScroll = isOldScroll;
			*bp = '\0';

			return bp - buf;
		}

		/*
		 * First check for the literal quote character.
		 * This accepts the following character as data
		 * no matter what it is.
		 */
		if (ch == wp->c_lnext)
		{
			ch = routine(ch);

			if (ch < 0)
			{
				wp->isScroll = isOldScroll;

				return bp - buf;
			}

			if (bp >= endBp)
			{
				/*
				 * Buffer is full.
				 */
				fputs("\a", stderr);
				fflush(stderr);

				continue;
			}

			*bp = ch;

			if (prompt)
				dpywrite(bp, 1);

			bp++;
			continue;
		}

		if (ch == wp->c_eof)
		{
			/*
			 * End of file
			 */
			wp->isScroll = isOldScroll;

			return bp - buf;
		}

		/*
		 * Single character erase.
		 * Because the distinction between backspace and rubout is
		 * so confused nowadays, makes things easier by accepting
		 * either of them if the erase character is one of them.
		 */
		if ((ch == wp->c_erase) ||
			(ISERASE(ch) && ISERASE(wp->c_erase)))
		{
			if (bp <= buf)
				continue;

			bp--;
			isRedrawNeeded = TRUE;

			continue;
		}

		if (ch == wp->c_werase)
		{
			/*
			 * Word erase.
			 */
			if (bp <= buf)
				continue;

			while ((bp > buf) &&
				((bp[-1] == '\n') || ISBLANK(bp[-1])))
			{
				bp--;
			}

			while ((bp > buf) &&
				(bp[-1] != '\n') && !ISBLANK(bp[-1]))
			{
				bp--;
			}

			isRedrawNeeded = TRUE;

			continue;
		}

		if (ch == wp->c_kill)
		{
			/*
			 * Line erase.
			 */
			if (bp <= buf)
				continue;

			if (bp[-1] == '\n')
				bp--;

			while ((bp > buf) && (bp[-1] != '\n'))
				bp--;

			isRedrawNeeded = TRUE;

			continue;
		}

		if (ch == wp->c_rprnt)
		{
			/*
			 * Retype line.
			 */
			if (prompt)
				dpyredraw();

			continue;
		}

		/*
		 * The character is not a special one, so just
		 * insert it into the buffer if there is room.
		 */
		if (bp >= endBp)
		{
			fputs("\a", stderr);
			fflush(stderr);

			continue;
		}

		*bp = ch;

		if (prompt)
			dpywrite(bp, 1);

		bp++;
	}
}


/*
 * Local routine to read until end of line character is reached.
 * The end of line is not returned in the buffer.
 * Notice: Errors and end of files are not treated well.
 */
static int
readLine(int oldCh)
{
	int newCh;

	do
	{
		newCh = dpyinputchar();
	}
	while (newCh == DPYINTR);

	if ((newCh < 0) || (newCh == '\n') || (newCh == '\r'))
		return -1;

	return newCh;
}

/* END CODE */
